/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


/*************************************************************************
 * File : bt_delete.c
 *
 * Modules:
 * BT_Delete()
 * BT_DeleteR()
 * BT_Redistribute_INode()
 * BT_Merge_INode()
 * BT_DeleteLeaf()
 * BT_RedistAndMergeLeaf()
 * BT_MergeLNode()
 * BT_CheckOverflowNodes()
 * BT_SelectSiblings()
 * BT_isEqualEntry()
 * BT_deleteEntries()
 * BT_Traverse()
 * error()
 *
 * Version 1.1
 * Coded by:  Panos Kalnis
 *************************************************************************/
#include "btree.h"
#include "bt_io.h"
#include "bt_node.h"
#include "bt_delete.h"

int totalDeleted = 0;


/***************************************************************************
 * BT_Delete(): 
 *   Routine to delete a record in the index
 *	If rd==NULL it deletes all records whith the given key.
 *   If rd!=NULL it deletes a record only when key AND the entire rd match
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   The number of deleted records
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_Delete(BTREE *btree, 		// the btree to delete from
	      RECORD_TYPE *rd,	// the record to delete
	      const KEYTYPE key	// the key to delete
	      ) {
  totalDeleted = 0;
	
  BT_DeleteR(btree, btree->Root, -1L, -1L, NULL, NULL, rd, key);
	
  return totalDeleted;
}



/***************************************************************************
 * BT_DeleteR(): 
 *   Recursive routine to implement delete
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   DEL_RESULT
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
DEL_RESULT BT_DeleteR(BTREE *btree,	// the btree to delete from
		      PTR currNodePtr,		// the offset of the current node
		      PTR lSibling,			// offset of left sibling of the current node
		      PTR rSibling,			// offset of right sibling of the current node
		      KEYTYPE *lKey,		// the key (at the parent level) which is left of the current node
		      KEYTYPE *rKey,		// the key (at the parent level) which is right of the current node
		      RECORD_TYPE *rd,		// the record to delete
		      const KEYTYPE key		// the key to delete
		      ) {
  DEL_RESULT result = NO_OP;
  DEL_RESULT opResult;
  KEYTYPE upKey;
	
  INODE *currNode;
  LNODE *newLNode;
  PTR child, lChildSibling, rChildSibling;
  PTR rightNodePtr;
  int childIndex;
  int i;
	
  INODE *sibNodeL;
  INODE *sibNodeR;
  int manyCopy;
  KEYTYPE *currLKey = NULL;
  KEYTYPE *currRKey = NULL;
	
  if (currNodePtr < 0L) error("Negative offset");

	
  BT_NewINode(&currNode);
	
  if (!BT_IOReadINode(btree, currNodePtr, currNode))  error("Cannot read INode");
		
  if (currNode->Level > 0) {  // not a leaf level INODE 
    child = BT_SelectSiblings(currNode, key, &lChildSibling, &rChildSibling, &childIndex);
    if (childIndex>0) currLKey = &(currNode->Keys[childIndex-1]); 
    if (childIndex<currNode->NumKey) currRKey = &(currNode->Keys[childIndex]);
    opResult = BT_DeleteR(btree, child, lChildSibling, rChildSibling, currLKey, currRKey, rd, key);
		
		
		
    if (opResult==REDIST_LEFT || opResult==REDIST_RIGHT) {
      BT_IOWriteINode(btree, currNodePtr, currNode);
    }
    else if (opResult==MERGE_LEFT || opResult==MERGE_RIGHT) {
      if (currNodePtr == btree->Root && currNode->NumKey == 1) { // root with only 1 key
					
	// remove root
	(btree->top)++;            
	if (btree->top > MAX_STACK) error("Stack size is too small !!");
	btree->Stacks[btree->top]=rightNodePtr;
				
	btree->Root = currNode->Children[0];
				
	printf("Root deleted! New root offset=%d, Level=%d\n", btree->Root, currNode->Level-1);
      }
      else {
	if (opResult==MERGE_RIGHT) childIndex++;
	for (i=childIndex; i<currNode->NumKey; i++) currNode->Children[i] = currNode->Children[i+1];
	for (i=childIndex-1; i<currNode->NumKey-1; i++) currNode->Keys[i] = currNode->Keys[i+1];
	currNode->NumKey--;
	BT_IOWriteINode(btree, currNodePtr, currNode);
      }
    }
  } 
  else {  // a leaf level INODE , the next level is LNODE
    child = BT_SelectSiblings(currNode, key, &lChildSibling, &rChildSibling, &childIndex);
    opResult = BT_DeleteLeaf(btree, child, lChildSibling, rChildSibling, &upKey, rd, key);
		
    if (opResult==REDIST_LEFT) currNode->Keys[childIndex-1] = upKey;
    if (opResult==REDIST_RIGHT) currNode->Keys[childIndex] = upKey;
			
    if (opResult==MERGE_LEFT) rightNodePtr = lChildSibling;
    else if (opResult==MERGE_RIGHT) {
      childIndex++;
      rightNodePtr = child;
    }

    if (opResult==MERGE_LEFT || opResult==MERGE_RIGHT) {
      if (currNodePtr == btree->Root && currNode->NumKey == 1) { // root with only 1 key
	currNode->Children[1] = rightNodePtr;
	currNode->Keys[0] = upKey;
	BT_NewLNode(&newLNode);
	newLNode->Next = rightNodePtr;
	btree->Leaf_Head = BT_IOWriteNewLNode(btree, newLNode);
	currNode->Children[0] = btree->Leaf_Head; 
	free(newLNode);
      }
      else {
	for (i=childIndex; i<currNode->NumKey; i++) currNode->Children[i] = currNode->Children[i+1];
	for (i=childIndex-1; i<currNode->NumKey-1; i++) currNode->Keys[i] = currNode->Keys[i+1];
	currNode->NumKey--;
      }
    }
			
    if (opResult!=NO_OP) {
      BT_IOWriteINode(btree, currNodePtr, currNode);
    }
			
  }
	
  if (opResult!=NO_OP && currNode->NumKey < (I_FAN_OUT-1)/2 && currNodePtr != btree->Root) {
    BT_NewINode(&sibNodeL);
    BT_NewINode(&sibNodeR);
		
    if (lSibling>=0 && !BT_IOReadINode(btree, lSibling, sibNodeL)) error("Cannot read INode");
    if (rSibling>=0 && !BT_IOReadINode(btree, rSibling, sibNodeR)) error("Cannot read INode");
		
    result = BT_Redistribute_INode(btree, currNode, sibNodeL, sibNodeR, currNodePtr, lSibling, rSibling, lKey, rKey);
		
    if (result==NO_OP) {
      if (lSibling>=0 && sibNodeL->NumKey + currNode->NumKey + 1 <= I_FAN_OUT - 1) { // merge with left
	BT_Merge_INode(btree, sibNodeL, currNode, lSibling, currNodePtr, *lKey);
	result = MERGE_LEFT;
      }
      else if (rSibling>=0 && sibNodeR->NumKey + currNode->NumKey + 1 <= I_FAN_OUT - 1) { // merge with right
	BT_Merge_INode(btree, currNode, sibNodeR, currNodePtr, rSibling, *rKey);
	result = MERGE_RIGHT;	
      }
    }	
    free(sibNodeL);
    free(sibNodeR);
  }

  free(currNode);
		
  return (result);
}



/***************************************************************************
 * BT_Merge_INode(): 
 *   Routine to merge two INODEs
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   All data will be placed in leftNode.
 *	leftNode is writen in the BTree file.
 *	rightNode is removed
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_Merge_INode(BTREE *btree,		// the btree to delete from
		   INODE *leftNode,	// the left one of the two nodes to be merged
		   INODE *rightNode,	// the right one of the two nodes to be merged
		   PTR leftNodePtr,	// the offset of the left node
		   PTR rightNodePtr,	// the offset of the right node
		   KEYTYPE parentKey	// the key (at the parent level) which is between the left and the right node
		   ) {
  int i;
	
  leftNode->Keys[leftNode->NumKey] = parentKey;
  leftNode->NumKey++;
	
  for (i=0; i<rightNode->NumKey; i++) leftNode->Keys[leftNode->NumKey + i] = rightNode->Keys[i];
	
  for (i=0; i<rightNode->NumKey+1; i++) leftNode->Children[leftNode->NumKey + i] = rightNode->Children[i];
	
  leftNode->NumKey += rightNode->NumKey;
  rightNode->NumKey = 0;
	
  (btree->top)++;            
  if (btree->top > MAX_STACK) error("Stack size is too small !!");
  btree->Stacks[btree->top]=rightNodePtr;
	
  BT_IOWriteINode(btree, leftNodePtr, leftNode);
}


/***************************************************************************
 * BT_Redistribute_INode(): 
 *   Routine to redistribute entries between two INODEs
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   DEL_RESULT
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   All data will be placed in leftNode.
 *	leftNode is writen in the BTree file.
 *	rightNode is removed
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
DEL_RESULT BT_Redistribute_INode(BTREE *btree,	// the btree to delete from
				 INODE *currNode,// The current node
				 INODE *sibNodeL,// The left sibling of the current node
				 INODE *sibNodeR,// The right sibling of the current node
				 PTR currNodePtr,// Offset of the current node
				 PTR	lSibling,	// Offset of the left sibling
				 PTR rSibling,	// Offset of the right sibling
				 KEYTYPE *lKey,	// the key (at the parent level) which is left of the current node
				 KEYTYPE *rKey	// the key (at the parent level) which is right of the current node
				 ) {
  DEL_RESULT result = NO_OP;
  int manyCopy;
  int i;
	
  if (lSibling>=0 && (sibNodeL->NumKey + currNode->NumKey >= I_FAN_OUT-1)) { // redistribute left
    manyCopy = (sibNodeL->NumKey + currNode->NumKey)/2 - currNode->NumKey;
			
    for (i=I_FAN_OUT-1; i>=manyCopy; i--) currNode->Children[i] = currNode->Children[i-manyCopy];
    for (i=0; i<manyCopy; i++) currNode->Children[i] = sibNodeL->Children[sibNodeL->NumKey+1 - manyCopy + i];
			
    for (i=I_FAN_OUT-2; i>=manyCopy; i--) currNode->Keys[i] = currNode->Keys[i-manyCopy];
    for (i=0; i<manyCopy-1; i++) currNode->Keys[i] = sibNodeL->Keys[sibNodeL->NumKey - manyCopy + i + 1];
			
    currNode->Keys[manyCopy-1] = *lKey;
    *lKey = sibNodeL->Keys[sibNodeL->NumKey - manyCopy];
			
    currNode->NumKey += manyCopy;
    sibNodeL->NumKey -= manyCopy;
			
    BT_IOWriteINode(btree, currNodePtr, currNode);
    BT_IOWriteINode(btree, lSibling, sibNodeL);
				
    result = REDIST_LEFT;
  }
  else if (rSibling>=0 && (sibNodeR->NumKey + currNode->NumKey >= I_FAN_OUT-1)) { // redistribute right

    manyCopy = (sibNodeR->NumKey + currNode->NumKey)/2 - currNode->NumKey;
			
    currNode->Keys[currNode->NumKey] = *rKey;
    *rKey = sibNodeR->Keys[manyCopy-1];
			
    for (i=0; i<manyCopy; i++) currNode->Children[currNode->NumKey + i+1] = sibNodeR->Children[i];
    for (i=0; i<sibNodeR->NumKey+1-manyCopy; i++) sibNodeR->Children[i] = sibNodeR->Children[i+manyCopy];
			
    for (i=1; i<manyCopy; i++) currNode->Keys[currNode->NumKey + i] = sibNodeR->Keys[i-1];
    for (i=0; i<sibNodeR->NumKey-manyCopy; i++) sibNodeR->Keys[i] = sibNodeR->Keys[i+manyCopy];
			
    currNode->NumKey += manyCopy;
    sibNodeR->NumKey -= manyCopy;
			
    BT_IOWriteINode(btree, currNodePtr, currNode);
    BT_IOWriteINode(btree, rSibling, sibNodeR);
			
    result = REDIST_RIGHT;
  }
		
  return result;
}





/***************************************************************************
 * BT_DeleteLeaf():
 *	Deletes all the matching records at the leaf level
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   DEL_RESULT
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   If the current node becomes empty, it is removed
 *	upKey contains the new key for the parent level
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
DEL_RESULT BT_DeleteLeaf(BTREE *btree,	// the btree to delete from
			 PTR currNodePtr,		// offset of the current node
			 PTR lSibling,			// offset of left sibling
			 PTR rSibling,			// offset of right sibling
			 KEYTYPE *upKey,		// the new key for the parent level
			 RECORD_TYPE *rd,		// the record to delete
			 const KEYTYPE key		// the key to delete
			 ) {
  DEL_RESULT result = NO_OP;
	
  LNODE *currNode = NULL;
  int checkONodes = FALSE;

	
  BT_NewLNode(&currNode);

  if (!BT_IOReadLNode(btree, currNodePtr, currNode)) error("Cannot read LNode");
	
  if (currNode->Entries[currNode->NumEntry].Key == key && currNode->overflowNum > 0) checkONodes = TRUE;
	
  // delete from LNODE
  checkONodes = checkONodes || BT_deleteEntries(currNode->Entries, &(currNode->NumEntry), rd, key);
	
	
  if (checkONodes) {
    BT_CheckOverflowNodes(btree, currNode, rd, key, TRUE);
  }	
	
  if (currNode->NumEntry < L_FAN_OUT/2) {
    result = BT_RedistAndMergeLeaf(btree, currNode, lSibling, rSibling, upKey);
  }

  if (result!=NO_OP && currNode->NumEntry==0) { 
    // remove node
    (btree->top)++;            
    if (btree->top > MAX_STACK) error("Stack size is too small !!");
    btree->Stacks[btree->top]=currNodePtr;
  }
  else {
    BT_IOWriteLNode(btree, currNodePtr, currNode);
  }

  free(currNode);
		
  return result;
}



/***************************************************************************
 * BT_RedistAndMergeLeaf():
 *	First it will try to redistribute entries with the siblings.
 *	If this fails, it will try to merge the current node with one of its siblings
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   DEL_RESULT
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	upkey contains the new key for the parent level
 *	The updates sibling is writen in the Btree file
 *
 * WARNINGS:
 *   The resulting LNODE may contain fewer that L_FAN_OUT/2 entries
 *	if it is not possible to redistribute or merge
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
DEL_RESULT BT_RedistAndMergeLeaf(BTREE *btree,	// the btree to delete from
				 LNODE *currNode,		// the current node
				 PTR lSibling,			// offset of the left sibling
				 PTR rSibling,			// offset of the right sibling
				 KEYTYPE *upKey			// the new key for the parent level
				 ) {
  LNODE *sibNodeR = NULL;
  LNODE *sibNodeL = NULL;
  int manyCopyR = 0;
  int manyCopyL = 0;
  int freeL = -1;
  int freeR = -1;
  int iLo, iHi, maxCopy, minCopy;
  DEL_RESULT result = NO_OP;
	
  BT_NewLNode(&sibNodeR);
  BT_NewLNode(&sibNodeL);
	
	
  // too few entries; try to redistribute with right sibling
  if (currNode->NumEntry < L_FAN_OUT/2 && rSibling>=0) {
    if (!BT_IOReadLNode(btree, rSibling, sibNodeR)) error("Cannot read LNode");
    freeR = L_FAN_OUT - sibNodeR->NumEntry; 
					
    if (currNode->NumEntry + sibNodeR->NumEntry >= L_FAN_OUT) {
      manyCopyR = (currNode->NumEntry + sibNodeR->NumEntry)/2 - currNode->NumEntry;
			
      // all dublicate entries must go to the same leaf node
      iLo = iHi = manyCopyR - 1;
      while (iLo>0 && sibNodeR->Entries[iLo].Key==sibNodeR->Entries[iLo-1].Key) iLo--;
      while (iHi<sibNodeR->NumEntry-1 && sibNodeR->Entries[iHi].Key==sibNodeR->Entries[iHi+1].Key) iHi++;
      minCopy = iLo;
      maxCopy = iHi + 1;
      if (maxCopy==L_FAN_OUT) maxCopy += sibNodeR->overflowNum;
			
      if (sibNodeR->NumEntry+sibNodeR->overflowNum-maxCopy >= L_FAN_OUT/2 
	  && currNode->NumEntry+maxCopy <= L_FAN_OUT) manyCopyR=maxCopy;
      else manyCopyR = minCopy;
    }
  }
	
  // still too few entries; try to redistribute with left sibling
  if (currNode->NumEntry+manyCopyR < L_FAN_OUT/2 && lSibling>=0) {
    if (!BT_IOReadLNode(btree, lSibling, sibNodeL)) error("Cannot read LNode");
    freeL = L_FAN_OUT - sibNodeL->NumEntry;

    // !!! make sure that the left sibling does not have overflow nodes
    if ((sibNodeL->Overflow < 0) && (currNode->NumEntry + sibNodeL->NumEntry >= L_FAN_OUT)) {
      manyCopyL = (currNode->NumEntry + sibNodeL->NumEntry)/2 - currNode->NumEntry;
			
      // all dublicate entries must go to the same leaf node
      iLo = iHi = sibNodeL->NumEntry - manyCopyL;
      while (iLo>0 && sibNodeL->Entries[iLo].Key==sibNodeL->Entries[iLo-1].Key) iLo--;
      while (iHi<sibNodeL->NumEntry-1 && sibNodeL->Entries[iHi].Key==sibNodeL->Entries[iHi+1].Key) iHi++;
      maxCopy = sibNodeL->NumEntry - iLo;
      minCopy = sibNodeL->NumEntry - iHi - 1;
			
      if (sibNodeL->NumEntry-maxCopy >= L_FAN_OUT/2 && currNode->NumEntry+maxCopy <= L_FAN_OUT) manyCopyL=maxCopy;
      else manyCopyL = minCopy;
    }
  }
	
  if (manyCopyR>=manyCopyL && manyCopyR>0) { // redistribute with right sibling
    memcpy(&(currNode->Entries[currNode->NumEntry]), &(sibNodeR->Entries[0]), manyCopyR*sizeof(LENTRY));
    memmove(&(sibNodeR->Entries[0]), &(sibNodeR->Entries[manyCopyR]), (L_FAN_OUT-manyCopyR)*sizeof(LENTRY));
    currNode->NumEntry += manyCopyR;
    sibNodeR->NumEntry -= manyCopyR;
			
    BT_CheckOverflowNodes(btree, sibNodeR, NULL, 0, FALSE);
			
    BT_IOWriteLNode(btree, rSibling, sibNodeR);
	
    *upKey = sibNodeR->Entries[0].Key;
    result = REDIST_RIGHT;
  }
  else if (manyCopyL>0) { // redistribute with left sibling
    memmove(&(currNode->Entries[manyCopyL]), &(currNode->Entries[0]), (L_FAN_OUT-manyCopyL)*sizeof(LENTRY));
    memcpy(&(currNode->Entries[0]), &(sibNodeL->Entries[sibNodeL->NumEntry - manyCopyL]), manyCopyL*sizeof(LENTRY));
    currNode->NumEntry += manyCopyL;
    sibNodeL->NumEntry -= manyCopyL;
			
    BT_IOWriteLNode(btree, lSibling, sibNodeL);
			
    *upKey = currNode->Entries[0].Key;
    result = REDIST_LEFT;
  }
  else { // try to merge
    if (freeL==-1 && lSibling>=0) {
      if (!BT_IOReadLNode(btree, lSibling, sibNodeL)) error("Cannot read LNode");
      freeL = L_FAN_OUT - sibNodeL->NumEntry;
    }
    if (freeR==-1 && rSibling>=0) {
      if (!BT_IOReadLNode(btree, rSibling, sibNodeR)) error("Cannot read LNode");
      freeR = L_FAN_OUT - sibNodeR->NumEntry;
    }
		
    if (freeL >= freeR && freeL >= currNode->NumEntry && sibNodeL->NumEntry>0) { // merge with left
      *upKey = BT_MergeLNode(btree, sibNodeL, currNode);
      BT_IOWriteLNode(btree, lSibling, sibNodeL);
			
      result = MERGE_LEFT;
    }
    else if (freeR >= currNode->NumEntry) { // merge with right
      *upKey = BT_MergeLNode(btree, currNode, sibNodeR);
			
      // remove node
      (btree->top)++;            
      if (btree->top > MAX_STACK) error("Stack size is too small !!");
      btree->Stacks[btree->top]=rSibling;
			
      result = MERGE_RIGHT;
    }
  }
	
  free(sibNodeR);
  free(sibNodeL);
	
  return result;
}


/***************************************************************************
 * BT_MergeLNode():
 *	Routine to merge two LNodes
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   The new key for the parent level
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	All entries are placed in the leftNode
 *
 * WARNINGS:
 *	This version does not maintain the Prev pointers in the LNODEs.
 *	To maintain the Prev pointers remove the commented part of the code
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
KEYTYPE BT_MergeLNode(BTREE *btree,		// the btree to delete from
		      LNODE *leftNode,	// the left one of the two nodes to be merged 
		      LNODE *rightNode	// the right one of the two nodes to be merged
		      ) {
  LNODE *tmpNode;
	
  BT_NewLNode(&tmpNode);
	
  memcpy(&(leftNode->Entries[leftNode->NumEntry]), &(rightNode->Entries[0]), rightNode->NumEntry*sizeof(LENTRY));
  leftNode->NumEntry += rightNode->NumEntry;
  rightNode->NumEntry = 0;
	
  leftNode->Next = rightNode->Next;
	
  // the following commented code takes care of the prev pointer
  /*
    if (rightNode->Next >=0) {
    if (!BT_IOReadLNode(btree, rightNode->Next, tmpNode)) error("Cannot read LNode");
    tmpNode->Prev = rightNode->Prev;
    BT_IOWriteLNode(btree, rightNode->Next, tmpNode);
    }
  */
	
  free(tmpNode);
	
  return leftNode->Entries[0].Key;
}


/***************************************************************************
 * BT_CheckOverflowNodes():
 *	If doDelete==TRUE this routine deletes a record from the overflow nodes
 *		and packs the remaining entries in the LNODE
 *	If doDelete==FALSE it only packs the entries from the overflow node
 *		in the LNODE. This is useful if the LNODE gives some of its entries
 *		to a sibling during redistribution
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   Node.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Emtpty overflow nodes are removed
 *
 * WARNINGS:
 *	None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_CheckOverflowNodes(BTREE *btree,			// the btree to delete from
			  LNODE *currNode,	// the current node
			  RECORD_TYPE *rd,	// the record to delete
			  const KEYTYPE key,	// the key to delete
			  int doDelete		// boolean value to control if the routine needs to delete any records
			  ) {
  ONODE *oNode = NULL;
  PTR overPtr;

  LENTRY *entryBuffer;
  int manyBuffer = 0;
  int bufferHead = 0;
  int manyCopy;
	
  if (currNode->Overflow < 0) return TRUE; // no overflow nodes
		
  BT_NewONode(&oNode);
		
  if (!(entryBuffer = (LENTRY *) malloc(currNode->overflowNum * sizeof(LENTRY)))) {
    error("Cannot allocate memory for overflow buffers");
  }
		
  overPtr = currNode->Overflow;
  while (overPtr>=0) {
    if (!BT_IOReadONode(btree, overPtr, oNode)) error("Cannot read ONode");
			
    if (doDelete) BT_deleteEntries(oNode->Entries, &(oNode->NumEntry), rd, key);
    memcpy(&(entryBuffer[manyBuffer]), &(oNode->Entries[0]), oNode->NumEntry * sizeof(LENTRY));
    manyBuffer += oNode->NumEntry;
			
    // remove overflow node
    (btree->top)++;            
    if (btree->top > MAX_STACK) error("Stack size is too small !!");
    btree->Stacks[btree->top]=overPtr;

    overPtr = oNode->Next;
  }
		
  // copy from buffer to LNODE	
  manyCopy = (L_FAN_OUT - currNode->NumEntry);  
  if (manyCopy > manyBuffer) manyCopy = manyBuffer;
  memcpy(&(currNode->Entries[currNode->NumEntry]), &(entryBuffer[bufferHead]), manyCopy * sizeof(LENTRY));
  currNode->NumEntry += manyCopy;
  bufferHead += manyCopy;
  currNode->overflowNum = manyBuffer-bufferHead;

  // write overflow nodes
  overPtr = -1;
  while (manyBuffer-bufferHead > 0) {
    // copy from buffer to ONODE	
    manyCopy = (manyBuffer-bufferHead) % O_FAN_OUT;
    if (manyCopy==0) manyCopy = O_FAN_OUT;  
    memcpy(&(oNode->Entries[0]), &(entryBuffer[manyBuffer-manyCopy]), manyCopy * sizeof(LENTRY));
    oNode->NumEntry = manyCopy;
    oNode->Next = overPtr;
    manyBuffer -= manyCopy;

    overPtr = BT_IOWriteNewONode(btree, oNode);
  }
		
  currNode->Overflow = overPtr;
		
  free(entryBuffer);
  free(oNode);

  return TRUE;
}



/***************************************************************************
 * BT_SelectSiblings():
 *   Routine to select a child according to the key value s.t. the node 
 *   pointed by the child would contain the key value
 *   It will choose the first child it hits even there might be more 
 *   qualifying children
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   1. The offset of the child.
 *	2. lSibling contains the offset of the child's left sibling
 *	3. lSibling contains the offset of the child's right sibling
 *	4. childIndex contains the offset of the child pointer inside the INODE->Children array
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Node.
 *
 * WARNINGS:
 *	Siblings must have the same parent. We CANNOT use the prev, next pointers.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
PTR BT_SelectSiblings(INODE *node,		// the INODE to be searched 
		      KEYTYPE key,		// the search key
		      PTR *lSibling,		// offset of the child's left sibling
		      PTR *rSibling,		// offset of the child's right sibling
		      int *childIndex	// offset of child pointer inside the Children array
		      ) {
  int i;
  int result = -1;

  if (node->NumKey == 0) {
    return (-1L);
  } 
	
  if (key < node->Keys[0]) {
    result = 0;
  }
  else {
    for (i=0; i<node->NumKey-1; i++) {
      if ((key >= node->Keys[i] && key<node->Keys[i+1]) ||
	  (key == node->Keys[i] && key<=node->Keys[i+1])) {
	result = i+1;
	break;
      }
    }
    if (result==-1) result = node->NumKey;
  }
	
  *lSibling = -1L;
  if (result>0) *lSibling = node->Children[result-1];
	
  *rSibling = -1L;
  if (result < node->NumKey) *rSibling = node->Children[result+1];
	
  *childIndex = result;
	
  return node->Children[result];
}


/***************************************************************************
 * BT_isEqualEntry():
 *	Checks if a leaf-level entry is equal to a (key, rd) pair.
 *	If rd==NULL only the key is checked for match (i.e., rd is ignored)
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *	TRUE if equal, FALSE otherwise
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Node.
 *
 * WARNINGS:
 *	None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_isEqualEntry(LENTRY *entry,		// leaf-level entry 
		    RECORD_TYPE *rd,	// the record to check
		    const KEYTYPE key	// the key to check
		    ) {
  int result = FALSE;
  int k;
	
  if (entry->Key == key) {
    result=TRUE;
    if (rd!=NULL) {		// the entire record must match
      for (k=0;k<D;k++) {
	if (entry->ver[k] != rd->fields[k]) {
	  result=FALSE;
	  break;
	}
      }
    }
  }
		
  return result;
}


/***************************************************************************
 * BT_deleteEntries():
 * 	Deletes from an array of leaf-level entries, these entries
 *	which match the (key, rd) pair.
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *	TRUE if equal, FALSE otherwise
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Node.
 *
 * WARNINGS:
 *	If rd==NULL only the key is checked for match (i.e., rd is ignored)
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_deleteEntries(LENTRY *entries,	// array of leaf-level entries to delete from
		     int *numEntry, 		// size of the entries array
		     RECORD_TYPE *rd, 	// the record to delete
		     const KEYTYPE key	// the key to delete
		     ) {
  int head, tail;
  int manyDeleted = 0;

  head = -1;
  tail = 0;
  while (tail < (*numEntry)) {
    if (BT_isEqualEntry(&(entries[tail]), rd, key) ) { // delete this entry = just skip
      manyDeleted++;
      totalDeleted++;
    }  
    else {
      head++;
      if (head!=tail) {
	memcpy(&(entries[head]), &(entries[tail]), sizeof(LENTRY));
      }
    }
    tail++; 
  }
  (*numEntry) = head + 1;
	
  return manyDeleted;
}





/***************************************************************************
 * BT_Traverse():
 * 	Routine to traverse the tree and print the nodes.
 *	If showLNodes==FALSE, only INODES are printed.
 *	If showLNodes==TRUE, all nodes are printed
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *	None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Node.
 *
 * WARNINGS:
 *	The INODES are visited by using level-order traversal.
 *	LNODES, however, are printed as soon as they are reached
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
void BT_Traverse(BTREE *btree,	// the btree to print
		 PTR currNodePtr,	// offset of the current node
		 int showLNodes	// controls if LNODES should be printed
		 ) {
  int qsize = 10000;
  int qhead = 0;
  int qtail = 0;
  PTR pageq[qsize];
	
  INODE *currNode;
  LNODE *currLNode;
  PTR child;
	
  int i; 

  if (currNodePtr < 0L) return;

  BT_NewINode(&currNode);
  BT_NewLNode(&currLNode);
  printf("\n\n");
	
  pageq[qtail] = currNodePtr;
  qtail++;
	
  while (qhead<qtail) {
    currNodePtr = pageq[qhead];
    qhead++;
		
    if (BT_IOReadINode(btree, currNodePtr, currNode)) {
      BT_PrintINode(currNode);
			
      if (currNode->Level > 0) {
	for (i=0; i<=currNode->NumKey; i++) {
	  pageq[qtail] = currNode->Children[i];
	  qtail++;
	  if (qtail==qsize) {
	    printf("Queue full - exiting...\n");
	    exit(0);
	  }
	}
      }
      else if (showLNodes) {
	for (i=0; i<=currNode->NumKey; i++) {
	  BT_IOReadLNode(btree, currNode->Children[i], currLNode);
	  BT_PrintLNode(currLNode);
	}
      }
    }
		
  }

  free(currNode);
  free(currLNode);
}

/***************************************************************************
 * error():
 * 	Routine to print an error message and exit the program
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *	None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *	Exits the program
 *
 * WARNINGS:
 *	None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
void error(char* errStr		// the error message
	   ) {
  printf(">>>> ERROR : %s\n", errStr);
  exit(-1);
}
