/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


/*************************************************************************
 *
 * File : btree.c
 *
 * Modules:
 * Routines to search in a B+ tree
 *	BT_Open()
 *	BT_Close()
 *	BT_Insert()
 *	BT_Search()
 *	BT_Locate()
 *	BT_SearchRange()
 *	refine()
 *	compact()
 *	BuildINodes()
 *	LoadINodes()
 *	IMBT_Locate()
 *	IMBT_LocateBaseINode()
 *	BT_RangeSearch_SortedIntervals()
 *	BT_RangeSearch_kNN()
 *	SearchRight() 
 *	SearchLeft()
 *	ProcessPoint()
 *	BT_DeleteKey()
 *	BT_DeleteRecord()
 *
 * Version 1.1
 * Coded by: Dai Haoyu		Modified by: Panos Kalnis, Rui Zhang
 *************************************************************************/
#include <float.h>

#include "btree.h"
#include "bt_io.h"
#include "bt_node.h"
#include "bt_search.h"
#include "bt_delete.h"
#include "bt_insert.h"
#include "offsetlist.h"
#include "io.h"
#include "customize.h"

PTR curleaf, tail;
extern DATA_TYPE min_r[NumRef], max_r[NumRef];

/***************************************************************************

* BT_Open():
*   Routine to open a BTree
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TURE if opened successfully else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The BTREE struct is properly initialsed
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Open(BTREE *btree,  /* the BTREE struct to be initialised */
            FILE *fp,      /* the file containing the B+ tree  */
            PTR  root,    /* the offset of the root node in the file */
            PTR  lhead    /* the offset of the leaf head node in the file */
	    ) {
  int result = TRUE;

  btree->BTreeFile = fp;
  btree->Root = root;
  btree->Leaf_Head = lhead;
  /**/
  btree->top=-1;

  return result;
}

/***************************************************************************

* functionName:
*   ShortDescription
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Close(BTREE *btree  /* the BTREE struct of which the B+ tree
                            * is to be closed */
	     ) {
  int result = FALSE;

  if (IOClose(btree->BTreeFile)){
    result = TRUE;
  }

  return result;
}

/***************************************************************************

* BT_Insert():
*   Routine to insert a new record to the B+ tree
*
* IMPORTS:
*   BT_NewINODE()
*   BT_IOWriteNewINode()
*   BT_IOWriteINode()
*   BT_IOReadINode()
*   BT_InsertR()
*   BT_INsertNewINode()
*
* RETURNS:
*   TRUE if insertion is successful else FALSE
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The B+ tree index file will be updated accordingly
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_Insert(BTREE *btree,      /* the btree to insert the record */
              const long recid,  /* the offset of record in the data file */
              //float key,
              const KEYTYPE key,  /* the key value of the reocord */
              RECORD_TYPE *rd
	      ) { 
  int result = FALSE;
  INODE *root,
    *newNode,
    *newRoot;
  PTR offset;
  KEYTYPE middleKey;  /*the key value to be moved up when splitting occurs*/

  if (!BT_NewINode(&root)) {
    printf("BT_Insert : Can't allocate memory for root node \n");
    return result;
  }

  if (btree->Root <0){  /* index is empty */
    root->Level = 0;
    root->NumKey = 0;
    btree->Root = BT_IOWriteNewINode(btree, root);
    printf("Root of b+ tree is created\n");
  }

  if (!BT_IOReadINode(btree, btree->Root, root)) {
    printf("BT_Insert : Can't read from index file\n");
    return result;
  }

  if (BT_InsertR(btree, recid, key, rd, root, &newNode, &middleKey)) {
    /* no split occurs */
    if (!BT_IOWriteINode(btree, btree->Root, root)) {
      printf("BT_Insert : Can't write to index file\n");
      return result;
    } else {
      free(root);
    }
  } else {  /* spliting occurs */
    printf("Splitting at root level\n");  

    if (!BT_VerifyINode(root)) {
      BT_PrintINode(root);
    }
    if (!BT_VerifyINode(newNode)) {
      BT_PrintINode(newNode);
    }
    BT_IOWriteINode(btree, btree->Root, root);
    BT_NewINode(&newRoot);
    newRoot->Level = root->Level + 1;
    newRoot->NumKey = 0;
    offset = BT_IOWriteNewINode(btree, newNode);
    /**/
    printf("************ Note: Offset of new root :  %d",btree->Root);

    BT_InsertNewINode(newRoot, btree->Root, middleKey, offset);
    btree->Root = BT_IOWriteNewINode(btree, newRoot);
    printf("offset of new root%d",btree->Root);

    free(root);
    free(newRoot);
    free(newNode);
  }

  result = TRUE;
  return result;
}



PTR BT_Locate( BTREE *btree,
               PTR currNodePtr,   /* the offset of the current node */          
               const KEYTYPE key  /* the key to be searched */    
	       ) {

  int i;
  INODE *currNode;
  PTR child, lchild;
  
  if (currNodePtr < 0L) {
    return -1;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
    if (currNode->Level > 0) {  /* not a leaf level INODE */
      child = BT_SelectChild(currNode, key);
      lchild = BT_Locate(btree, child, key);
    } else {  /* a leaf level INODE , the next level is LNODE */
      lchild = BT_SelectChild(currNode, key);      
    }
  }
  free(currNode);
  return (lchild);  

}


/***************************************************************************
 * BT_Search():
 *   Routine to search a record in the index
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   TRUE if searching is successful else FALSE.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   The set of record ids of found record will be passed as an offsetlist
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_Search(BTREE *btree,     /* the B+ tree to searched from */
              PTR currNodePtr,   /* the offset of the current node */
              RECORD_TYPE *rd,
              const KEYTYPE key  /* the key to be searched */
	      ) {
  int result = TRUE;
  int i;
  INODE *currNode;
  PTR child;
  OFFSETLIST *subList = NULL;

  if (currNodePtr < 0L) {
    return result;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
    if (currNode->Level > 0) {  /* not a leaf level INODE */

      /* alternative 2 : to invoke BT_Search only for the left-most qualifing
       *                 child and perform sequential scan at LNODE level */
      child = BT_SelectChild(currNode, key);
      BT_Search(btree, child, rd, key);

    } else {  /* a leaf level INODE , the next level is LNODE */
      /* alternative 2 */
      child = BT_SelectChild(currNode, key);
      BT_SearchLeaf(btree, child, rd, key,&subList);
      BT_InsertSrhResult(btree,subList);
      DeleteList(&subList);
      /* alternative 2 */
    }
  }

  free(currNode);
  return (result);
}



/***************************************************************************

* BT_SearchRange():
*   Routine to search a range in the index
*
* IMPORTS:
*   None.
*
* RETURNS:
*   TRUE if searching is successful else FALSE.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The set of record ids of found record will be passed as an offsetlist
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
int BT_SearchRange(BTREE *btree,     /* the B+ tree to searched from */
		   PTR currNodePtr,   /* the offset of the current node */
		   const KEYTYPE key1,  /* lower bound of the range */
		   const KEYTYPE key2,   /* upper bound of the range */
		   FLTRANGE *rg
		   ) {
  int result = TRUE;
  int i;
  INODE *currNode;
  OFFSETLIST *subList = NULL;
  PTR child;

  if (currNodePtr < 0L) {
    return result;
  }
  
  BT_NewINode(&currNode);

  if (BT_IOReadINode(btree, currNodePtr, currNode)){
    if (currNode->Level > 0) {  /* not a leaf level INODE */      

      /* alternative 2 : to invoke BT_Search only for the left-most qualifing
       *                 child and perform sequential scan at LNODE level */
      child = BT_SelectChild(currNode, key1);
      BT_SearchRange(btree, child, key1, key2, rg);

    } else {  /* a leaf level INODE , the next level is LNODE */
      /* alternative 2 */
      subList = NULL;
      child = BT_SelectChild(currNode, key1);
      BT_SearchLeafRange(btree, child, key1, key2, rg, &subList);
      BT_InsertSrhResult(btree, subList);
      DeleteList(&subList);
      /* alternative 2 */
    }
  }

  free(currNode);
  return (result);
}

/***************************************************************************
 * refine():
 *   Routine to add previus pointer between leafnodes.
 *
 * IMPORTS:
 *   Pointer to the b-tree.
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

void refine(BTREE *btree)   
{

  PTR pointer, prev; 
  LNODE *lNode=NULL;
  int Continue=1;
   
  BT_NewLNode(&lNode);        
  pointer=btree->Leaf_Head;
   
  prev=-1; 
  while (Continue) {       
    BT_IOReadLNode(btree, pointer, lNode);          
    lNode->Prev=prev;
    BT_IOWriteLNode(btree, pointer, lNode);
    prev=pointer;
    pointer=lNode->Next;
    if (pointer<0) {
      Continue=0;
    } else { 
    }
  }
  free(lNode);
}

/***************************************************************************
 * compact():
 *   Routine to compact the nodes of the b-tree to achieve high utilization rate.
 *
 * IMPORTS:
 *   Pointer to the b-tree.
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

void compact(BTREE *btree)
{
  PTR addr1, addr2, oriaddr1, fromover, toover, tonext, leafhead, treeroot; 
  LNODE *lNode1, *lNode2;
  ONODE *oNode1, *oNode2;
  int to, from, i, j, flag, tonodetype, fromnodetype;
  int clfanout;
  KEYTYPE x,y;
	
  if(compactrate<0.75)
    {
      printf("please specify a larger utilization rate.\n");
      return;
    }
	
  clfanout=(int)(compactrate * L_FAN_OUT);
	
  BT_NewLNode(&lNode1);
  BT_NewLNode(&lNode2);
  BT_NewONode(&oNode1);
  BT_NewONode(&oNode2);
  addr1=leafhead=btree->Leaf_Head;
  treeroot=btree->Root;
	
  //find the first node with less than clfanout entries
  do
    {
      BT_IOReadLNode(btree, addr1, lNode1);
      oriaddr1=addr1;
      addr1=lNode1->Next;
      to=lNode1->NumEntry;
    }while(to==clfanout-1 && addr1>=0);
	
  toover=lNode1->Overflow;
  tonext=addr2=addr1; //addr2 points to the next node of lNode1;
  from=0;
	
  flag=1;
  tonodetype=fromnodetype=1; //1 for leaf; 0 for overflow node.
	
  if(addr2>=0) BT_IOReadLNode(btree, addr2, lNode2);
  else flag=0;
	
  while (flag>0)
    {
      if(fromnodetype)
	{
	  for( ; to < clfanout && from < lNode2->NumEntry ; to++, from++)
	    lNode1->Entries[to]=lNode2->Entries[from];
	}
      else
	{
	  for( ; to < clfanout && from < oNode2->NumEntry ; to++, from++)
	    lNode1->Entries[to]=oNode2->Entries[from];
	}
		
      if(fromnodetype && from==lNode2->NumEntry)
	{
	  from=0;
	  fromover=lNode2->Overflow;
	  addr2=lNode2->Next;
			
	  if(fromover>=0)
	    {
	      BT_IOReadONode(btree,fromover,oNode2);
	      fromnodetype=0;
	    }
	  else if(addr2>=0)
	    {
	      BT_IOReadLNode(btree, addr2, lNode2);
	    }
	  else flag=0;
	}
      else if(!fromnodetype && from==oNode2->NumEntry)
	{
	  from=0;
	  fromover=oNode2->Next;
	  if(fromover>=0)
	    {
	      BT_IOReadONode(btree,fromover,oNode2);
	    }
	  else if(addr2>=0)
	    {
	      BT_IOReadLNode(btree, addr2, lNode2);
	      fromnodetype=1;
	    }
	  else flag=0;
	}
		
      if(to==clfanout)
	{
	  if(!flag) 
	    {
	      lNode1->NumEntry=to;
	      lNode1->Next=-1;
	      lNode1->Overflow=-1;
	      lNode1->overflowNum=0;
	      BT_IOWriteLNode(btree, oriaddr1, lNode1);
	      flag=-1;
	    }
			
	  else
	    {
	      if(toover>=0)
		{
		  lNode1->NumEntry=to;
		  lNode1->Next=toover;
		  lNode1->Overflow=-1;
		  lNode1->overflowNum=0;					
		  BT_IOWriteLNode(btree, oriaddr1, lNode1);

		  BT_IOReadONode(btree,toover,oNode1);
		  oriaddr1=toover;
		  toover=oNode1->Next;
		  tonodetype=0;
		}
	      else if(tonext>=0)
		{
		  lNode1->NumEntry=to;
		  lNode1->Next=tonext;
		  lNode1->Overflow=-1;
		  lNode1->overflowNum=0;					
		  BT_IOWriteLNode(btree, oriaddr1, lNode1);

		  BT_IOReadLNode(btree, tonext, lNode1);
		  oriaddr1=tonext;
		  toover=lNode1->Overflow;
		  tonext=lNode1->Next;
		  tonodetype=1;
		}
	      else flag=0;
				
	      to=0;
	    }
	}
    }
	
  if(flag != -1)
    {
      lNode1->NumEntry=to;
      lNode1->Next=-1;
      lNode1->Overflow=-1;
      lNode1->overflowNum=0;
      BT_IOWriteLNode(btree, oriaddr1, lNode1);
    }
	
  curleaf=btree->Leaf_Head;
  BuildINodes(treeroot, btree, &x, &y);

  free(lNode1);
  free(lNode2);
  free(oNode1);
  free(oNode2);
}

/***************************************************************************
 * BuildINodes():
 *   Routine called by compact(), to build the internal nodes after compacting
 *	the leaf nodes.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

int BuildINodes(PTR ptr, BTREE *btree, KEYTYPE *keymin, KEYTYPE *keymax)
{
  int i, j, flag;
  PTR p;
  int cifanout;
  INODE *pinode, *pchild;
  LNODE *lNode;
  KEYTYPE curkeymin, curkeymax, childkeymin, childkeymax;
	
  cifanout=(int)(compactrate * I_FAN_OUT);
	
  BT_NewINode(&pinode);
  BT_IOReadINode(btree, ptr, pinode);
	
  BT_NewINode(&pchild);
  BT_NewLNode(&lNode);

  flag=1;

  curkeymin=KEY_MAX;
  curkeymax=KEY_MIN;

  if(pinode->Level>0)
    {
      for(i=0;i<pinode->NumKey+1 && flag>0;i++)
	{
	  flag=BuildINodes(pinode->Children[i], btree, &childkeymin, &childkeymax);
	  if(flag>=0)
	    {
	      if(i==0) curkeymin=childkeymin;
	      else pinode->Keys[i-1]=childkeymin;		
	    }
	}
		
      *keymin=curkeymin;
      pinode->NumKey=i-1;
		
      BT_IOWriteINode(btree, ptr, pinode);
		
      free(pinode);
      free(pchild);
      free(lNode);

      if(i==1 && flag<0) return -1;//this inode has no child inode
      else if(flag==0) return 0;
      else return 1;					
    }
	
  if(pinode->Level==0)
    {
      if(curleaf>=0)
	{
	  BT_IOReadLNode(btree, curleaf, lNode);
	  pinode->Children[0]=curleaf;
	  pinode->NumKey=0;
	  curleaf=lNode->Next;
	  *keymin=lNode->Entries[0].Key;
	}
      else 
	{
	  free(pinode);
	  free(pchild);
	  free(lNode);
	  return -1;
	}
			
      for( i=0; curleaf>=0 && i<cifanout-1; i++ )
	{
	  BT_IOReadLNode(btree, curleaf, lNode);
	  pinode->Children[pinode->NumKey+1]=curleaf;
	  pinode->Keys[pinode->NumKey]=lNode->Entries[0].Key;
	  pinode->NumKey++;
	  curleaf=lNode->Next;
	}
		
      BT_IOWriteINode(btree, ptr, pinode);

      free(pinode);
      free(pchild);
      free(lNode);		
		
      if(curleaf<0) return 0;
      else return 1;
    }
}

/***************************************************************************
 * LoadINodes():
 *   Routine to load internal nodes.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

IMINODE* LoadINodes(PTR ptr, BTREE *btree)
{
  int i;
  INODE *pinode, *pbaseiminode;
  IMINODE *piminode;

	
  BT_NewINode(&pinode);
  BT_IOReadINode(btree, ptr, pinode);
	
  if(pinode->Level>0)
    {
      piminode=(IMINODE *)malloc(sizeof(IMINODE));
      piminode->Level=pinode->Level;
      piminode->NumKey=pinode->NumKey;
      for(i=0;i<pinode->NumKey;i++) piminode->Keys[i]=pinode->Keys[i];
      for(i=0;i<=pinode->NumKey;i++) 
	{piminode->Children[i]=LoadINodes(pinode->Children[i], btree);}
		
      free(pinode);

      return piminode;
    }
	
  else if(pinode->Level==0)
    {
      pbaseiminode=(INODE *)malloc(sizeof(INODE));
      pbaseiminode->Level=pinode->Level;
      pbaseiminode->NumKey=pinode->NumKey;
      for(i=0;i<pinode->NumKey;i++) pbaseiminode->Keys[i]=pinode->Keys[i];
      for(i=0;i<=pinode->NumKey;i++)
	{ pbaseiminode->Children[i]=pinode->Children[i];}
		
      free(pinode);
		
      return (IMINODE*)pbaseiminode;
    }
	
  else
    {
      free(pinode);
      printf("Load INodes Error\n");
      exit(0);
    }
}

/***************************************************************************
 * IMBT_Locate():
 *   For in-memory internal nodes, to locate the leaf node contains key.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

PTR IMBT_Locate(IMINODE *piminode, KEYTYPE key)
{
  int i;
  INODE *pinode;
  pinode=(INODE *)IMBT_LocateBaseINode(piminode,key);
	
  //scan the base inode to locate the page to be read
  if(pinode->NumKey==0) return -1;
  else
    {
      if(key<pinode->Keys[0])	return pinode->Children[0];
		
      for(i=0;i<pinode->NumKey-1;i++)
	{
	  if(pinode->Keys[i]<=key && key<=pinode->Keys[i+1]) return pinode->Children[i+1];
	}
		
      return pinode->Children[pinode->NumKey];
    }
  //End scan the base inode to locate the page to be read
}

/***************************************************************************
 * IMBT_LocateBaseINode():
 *   For in-memory internal nodes, to locate the leaf node contains key 
 *	this is for base level internal nodes, i.e., internal nodes with level=0.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

IMINODE* IMBT_LocateBaseINode(IMINODE *piminode, KEYTYPE key)
{
  int i;
	
  if(piminode->Level==0) return piminode;
	
  if(piminode->Level==1)
    {
      if(key<piminode->Keys[0])	return piminode->Children[0];
		
      for(i=0;i<piminode->NumKey-1;i++)
	{
	  if(piminode->Keys[i]<=key && key<piminode->Keys[i+1]) return piminode->Children[i+1];
	}
		
      return piminode->Children[piminode->NumKey];
    }
	
  if(piminode->Level>1)
    {
      if(key<piminode->Keys[0])	return IMBT_LocateBaseINode(piminode->Children[0],key);
		
      for(i=0;i<piminode->NumKey-1;i++)
	{
	  if(piminode->Keys[i]<=key && key<piminode->Keys[i+1]) return IMBT_LocateBaseINode(piminode->Children[i+1],key);
	}
		
      return IMBT_LocateBaseINode(piminode->Children[piminode->NumKey],key);
    }
}

/***************************************************************************
 * BT_RangeSearch_SortedIntervals():
 *   b-tree range search, optimized for a set of intervals sorted by
 *	their lower bounds.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

int BT_RangeSearch_SortedIntervals(BTREE *btree, RecordListNode *cddhead, Interval *intvlarray, int ni)
{
  int cur,i,j, conti;
  Interval intvl;
  PTR address, over;
  LNODE *lNode;
  ONODE *oNode;
  RecordListNode *pRLN;
	
  if(ni<=0) return 0;
	
  BT_NewLNode(&lNode);
  BT_NewONode(&oNode);
	
  address=IMBT_Locate(btree->imroot, intvlarray[0].lower);
  BT_IOReadLNode(btree,address,lNode);
	
  cur=i=0;
  conti=1;
  intvl=intvlarray[cur++];
  while(conti)
    {
      for(;i<lNode->NumEntry && lNode->Entries[i].Key < intvl.lower; i++);
      if(i<lNode->NumEntry)
	{
	  for(;i<lNode->NumEntry && lNode->Entries[i].Key <= intvl.higher ;i++)
	    {
	      pRLN=(RecordListNode *)malloc(sizeof(RecordListNode));
	      for(j=0;j<D;j++) pRLN->rcd.fields[j]=lNode->Entries[i].ver[j];
	      pRLN->rid=lNode->Entries[i].Offset;
	      pRLN->next=cddhead->next;
	      cddhead->next=pRLN;
	    }
	}
		
      if(i==lNode->NumEntry)
	{
	  over=lNode->Overflow;
	  while(over >= 0)
	    {
	      BT_IOReadONode(btree,over,oNode);
	      for(i=0;i<oNode->NumEntry && oNode->Entries[i].Key < intvl.lower; i++);
	      if(i<oNode->NumEntry)
		{
		  for(;i<oNode->NumEntry && oNode->Entries[i].Key<=intvl.higher;i++)
		    {
		      pRLN=(RecordListNode *)malloc(sizeof(RecordListNode));
		      for(j=0;j<D;j++) pRLN->rcd.fields[j]=oNode->Entries[i].ver[j];
		      pRLN->rid=oNode->Entries[i].Offset;
		      pRLN->next=cddhead->next;
		      cddhead->next=pRLN;
		    }
		}
	      over=oNode->Next;
	    }
			
	  if(lNode->Next >=0)
	    {
	      BT_IOReadLNode(btree,lNode->Next,lNode);
	      i=0;
	    }
	  else conti=0;
	}
		
      if(lNode->Entries[i].Key>intvl.higher)
	{
	  if(cur<ni)
	    {
	      intvl=intvlarray[cur++];
	      if(intvl.higher > lNode->Entries[lNode->NumEntry-1].Key)
		{
		  address=IMBT_Locate(btree->imroot, intvl.lower);
		  BT_IOReadLNode(btree,address,lNode);
		  i=0;
		}
	      else
		{
		  //for(;lNode->Entries[i].Key < intvl.lower; i++);
		}				
	    }
	  else conti=0;
	}
    }

  free(lNode);
  free(oNode);
}

KEYTYPE keylow[NumRef], keyhigh[NumRef];//keylow/keyhigh records the lower/higher key searched
PTR pagelow[NumRef], pagehigh[NumRef], donelow[NumRef+1], donehigh[NumRef+1]; //for BT_RangeSearch_kNN();
//pagelow/pagehigh records next page to search for lower/higher keys;
//donelow/donehigh records the last page searched for lower/higher keys;

/***************************************************************************
 * BT_RangeSearch_kNN():
 *   b-tree range search, optimized for kNN search. Maintaining the candidate list 
 *	is integrated into it to avoid passing back large number of candidates.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

int BT_RangeSearch_kNN(BTREE *btree, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo, IntervalListNode *intvlarray)
{
  int curpar, i;
  KEYTYPE curkeylow, curkeyhigh;
  PTR address, over;
  LNODE *lNode; 
  ONODE *oNode;
  IntervalListNode *pILN, *pnext;

  BT_NewLNode(&lNode);
  BT_NewONode(&oNode);
  for(curpar=0;curpar<NumRef;curpar++)
    {
      curkeylow=intvlarray[curpar*2].lower;
      curkeyhigh=intvlarray[curpar*2+1].higher;
      if(curkeylow>=0)
	{
	  if(keylow[curpar]<0) //this partition is not searched before
	    {
	      address=IMBT_Locate(btree->imroot, curkeylow);
	      BT_IOReadLNode(btree,address,lNode);
	      for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i], query, answerset, knninfo);
				
	      over=lNode->Overflow;
	      while(over >= 0)
		{
		  BT_IOReadONode(btree,over,oNode);
		  for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i], query, answerset, knninfo);
		  over=oNode->Next;
		}
				
	      if(lNode->Prev<0 || (curpar>0 && lNode->Prev ==donehigh[curpar-1]))
		{
		  keylow[curpar]=min_r[curpar];
		  donelow[curpar]=address;
		}
	      else
		{
		  keylow[curpar]=lNode->Entries[0].Key;
		  pagelow[curpar]=lNode->Prev;
		  donelow[curpar]=address;
		}
				
	      if(lNode->Next<0 || lNode->Next == donelow[curpar+1]) 
		{
		  keyhigh[curpar]=max_r[curpar];
		  donehigh[curpar]=address;
		}
	      else if(lNode->Entries[lNode->NumEntry-1].Key <= curkeyhigh)
		SearchRight(lNode->Next, curkeyhigh, btree, curpar, query, answerset, knninfo);
	      else
		{
		  keyhigh[curpar]=lNode->Entries[lNode->NumEntry-1].Key;
		  pagehigh[curpar]=lNode->Next;
		  donehigh[curpar]=address;
		}
	    }
			
	  else
	    {
	      if(curpar>0 && pagelow[curpar]==donehigh[curpar-1]) keylow[curpar]=min_r[curpar];
	      else if(keylow[curpar]>curkeylow)
		SearchLeft(pagelow[curpar], curkeylow, btree, curpar, query, answerset, knninfo);
				
	      if(pagehigh[curpar]==donelow[curpar+1]) keyhigh[curpar]=max_r[curpar];
	      else if(keyhigh[curpar]<curkeyhigh)
		SearchRight(pagehigh[curpar], curkeyhigh, btree, curpar, query, answerset, knninfo);
	    }
	}
    }
	
  free(intvlarray);
	
  free(lNode);
  free(oNode);

  if(knninfo->numanswer==knninfo->k && knninfo->farthest<=knninfo->r*knninfo->r) return 0;	
  else return 1;
}

/***************************************************************************
 * SearchLeft():
 *   Routine called by BT_RangeSearch_kNN() to search to the left neighbor leaf.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

void SearchLeft(PTR address, float curkeylow, BTREE *btree, int curpar, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo)
{
  int i, j;
  int continueflag;
  ONODE *oNode;
  LNODE *lNode;
  PTR over;

  if(address<0) return;

  BT_NewLNode(&lNode);   
  BT_NewONode(&oNode);
	
  continueflag=1;
	
  while(continueflag>0)
    {
      BT_IOReadLNode(btree,address,lNode);
		
      for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i], query, answerset, knninfo);
		
      over=lNode->Overflow;
		
      while(over >= 0)
	{
	  BT_IOReadONode(btree,over,oNode);
	  for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i], query, answerset, knninfo);
	  over=oNode->Next;
	}
		
      if(lNode->Entries[0].Key < curkeylow || lNode->Prev<0)
	{
	  continueflag=0;
	  donelow[curpar]=address;
	}
      else if(curpar>0 && lNode->Prev==donehigh[curpar-1])
	{
	  continueflag=-1;
	  keylow[curpar]=min_r[curpar];
	}
      else address = lNode->Prev;
    }
	
  if(continueflag==0)
    {
      keylow[curpar]=lNode->Entries[0].Key;
      pagelow[curpar]=lNode->Prev;
    }
  free(lNode);
  free(oNode);
}

/***************************************************************************
 * SearchRight():
 *   Routine called by BT_RangeSearch_kNN() to search to the right neighbor leaf.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

void SearchRight(PTR address, float curkeyhigh, BTREE *btree, int curpar, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo)
{
  int i, j;
  int continueflag;
  ONODE *oNode;
  LNODE *lNode;
  PTR over;
	
  if(address<0) return;

  BT_NewLNode(&lNode);   
  BT_NewONode(&oNode);
	
  continueflag=1;
	
  while(continueflag>0)
    {
      BT_IOReadLNode(btree,address,lNode);
		
      for(i=0;i<lNode->NumEntry;i++) ProcessPoint(lNode->Entries[i], query, answerset, knninfo);
		
      over=lNode->Overflow;
		
      while(over >= 0)
	{
	  BT_IOReadONode(btree,over,oNode);
	  for(i=0;i<oNode->NumEntry;i++) ProcessPoint(oNode->Entries[i], query, answerset, knninfo);
	  over=oNode->Next;
	}
		
      if(	lNode->Entries[lNode->NumEntry-1].Key > curkeyhigh || lNode->Next<0) 
	{
	  continueflag=0;
	  donehigh[curpar]=address;
	}
      else if(lNode->Next==donelow[curpar+1])
	{
	  continueflag=-1;
	  keyhigh[curpar]=max_r[curpar];
	}
      else address = lNode->Next;
    }
	
  if(continueflag==0)
    {
      keyhigh[curpar]=lNode->Entries[lNode->NumEntry-1].Key;
      pagehigh[curpar]=lNode->Next;
    }
  free(lNode);
  free(oNode);	
}

/***************************************************************************
 * ProcessPoint():
 *   Routine to process a point in kNN search.
 *
 * IMPORTS:
 *
 * RETURNS:
 *   None.
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None
 *
 * WARNINGS:
 *   None.
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/

#define CHECKSAME

void ProcessPoint(LENTRY le, RECORD_TYPE *query, RecordListNode *answerset, KnnInfo *knninfo)
{
  register int i, j;
  DATA_TYPE CurDist;
  register DATA_TYPE temp;
	
  for(i=CurDist=0;i<D;i++)
    {
      temp = le.ver[i]-query->fields[i];
      CurDist += temp*temp;
    }
	
  if(knninfo->numanswer<knninfo->k)
    {
		
#ifdef CHECKSAME
      for (i=0;i<knninfo->numanswer && answerset[i].rid!=le.Offset ;i++);
      if(i<knninfo->numanswer) return;
#endif		
      answerset[knninfo->numanswer].rid=le.Offset;
      knninfo->distance[knninfo->numanswer]=CurDist;
      if(CurDist>knninfo->farthest)
	{
	  knninfo->farthest=CurDist;
	  knninfo->farthestnum=knninfo->numanswer;
	}
      knninfo->numanswer++;
    }
	
  else
    {
      if(CurDist<knninfo->farthest)
	{
#ifdef CHECKSAME
	  for (i=0;i<knninfo->numanswer && answerset[i].rid!=le.Offset ;i++);
	  if(i<knninfo->numanswer) return;
#endif
	  answerset[knninfo->farthestnum].rid=le.Offset;
	  knninfo->distance[knninfo->farthestnum]=CurDist;
		
	  //find the farthest point and record it
	  knninfo->farthest=0;
	  for(j=0;j<knninfo->k;j++)
	    {
	      if(knninfo->distance[j]>knninfo->farthest)
		{
		  knninfo->farthest=knninfo->distance[j];
		  knninfo->farthestnum=j;
		}
	    }
	  //End find the farthest point and record it
	}
    }	
}

/***************************************************************************
 * BT_DeleteKey(): 
 *   Routine to delete all entries with a given key
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   The number of deleted records
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   Only the key is used to check for qualifying entries
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_DeleteKey(BTREE *btree,	// the btree to delete from
		 const KEYTYPE key	// the key to delete
		 ) {
  return BT_Delete(btree, NULL, key);
}

/***************************************************************************
 * BT_DeleteRecord(): 
 *   Routine to delete all entries which match a key and the record rd
 *
 * IMPORTS:
 *   None.
 *
 * RETURNS:
 *   The number of deleted records
 *
 * ERROR RETURNS:
 *   None.
 *
 * SIDE EFFECTS:
 *   None.
 *
 * WARNINGS:
 *   Both key and rd are used to check for qualifying entries
 *
 * BUGS:
 *   Not discovered yet
 ***************************************************************************/
int BT_DeleteRecord(BTREE *btree,	// the btree to delete from
		    RECORD_TYPE *rd,		// the record to delete
		    const KEYTYPE key		// the key to delete
		    ) {
  if (rd==NULL) {
    printf("Record to be deleted cannot be NULL\n");
    exit(0);
  }
	
  return BT_Delete(btree, rd, key);
}
