/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */



/*************************************************************************
 *
 * File : offsetlist.c
 *
 * Modules:
 * The routines to manipulate OFFSETLIST
 *	NewList()
 *	DeleteList()
 *	DupList()
 *	JoinList()
 *	InsertList()
 *	PrintList()
 *	PrintList_v2()
 *	IntersectList()
 *
 * Version 1.1
 * Coded by: Dai Haoyu
 *************************************************************************/
#include "btree.h"
#include "offsetlist.h"

/***************************************************************************

* NewList():
*   Routine to allocate memory space for a OFFSETLIST node
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   If the listNode points to a pointer which points to some memory space
*   ERROR message will be printed and no allocation will be performed
*
* SIDE EFFECTS:
*   The pointer pointed to by listNode will point to a OFFSETLIST node
*
* WARNINGS:
*   The argument passed in should point to a null pointer so as to avoid
*   possible tangling space.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void NewList(OFFSETLIST **listNode
	     ) {

  if (*listNode != NULL){
    printf("NewList(): *listNode is not NULL.\n");
    return;
  }else{
    *listNode = (OFFSETLIST *)malloc(sizeof(OFFSETLIST));
    (*listNode)->next = NULL;
  }
}


/***************************************************************************

* DeleteList():
*   Routine to repeatedly delete all nodes in an OFFSETLIST list
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   The memory space hold by all the nodes in the list are released. The head
*   pointer of the list is changed to NULL since it points to nowhere.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void DeleteList(OFFSETLIST **listHead  /* the pointer pointing to the head
                                        * pointer of the OFFSETLIST list  */
		) {
  OFFSETLIST *nextNode, *currNode;

  currNode = *listHead;
  while (currNode){
    nextNode = currNode->next;
    free(currNode);
    currNode = nextNode;
  }
  *listHead = NULL;
}

/***************************************************************************

* DupList():
*   Routine to duplicate an OFFSETLIST list
*
* IMPORTS:
*   void NewList()
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   ERROR message will be displayed if the origList points to nowhere
*   ERROR message will be displayed if the newList points to a pointer
*   that points to some memory space to avoid tangling space
*
* SIDE EFFECTS:
*   Memory space will be allocated as the list is duplicated.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void DupList(const OFFSETLIST *origList,  /* the list to be duplicated */
             OFFSETLIST **newList         /* the duplicated list */
	     ) {
  const OFFSETLIST *currNode;
  OFFSETLIST *newNode,
    *NewCurrNode;

  if (origList == NULL){
    /*
      printf("DupList : the original list is empty\n");
    */
    return;
  }
  if (*newList != NULL){
    printf("DupList : the newList points to some space\n");
    return;
  }

  currNode = origList;
  if((*newList = (OFFSETLIST *) malloc(sizeof(OFFSETLIST))) == NULL) {
    printf("DupList : Cannot allocate memory\n");
  }
  NewCurrNode = *newList;
  NewCurrNode->Offset = currNode->Offset;
  NewCurrNode->next = NULL;

  while (currNode->next){
    currNode = currNode->next;
    NewCurrNode->next = (OFFSETLIST *) malloc(sizeof(OFFSETLIST));
    if (NewCurrNode->next == NULL) {
      printf("DupList : Cannot allocate memory\n");
    }
    NewCurrNode = NewCurrNode->next;
    NewCurrNode->Offset = currNode->Offset;
    NewCurrNode->next = NULL;
  }
}

/***************************************************************************

* JoinList:
*   Routine to join two lists into one single list in increasing order
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   this routine does not allocate space for newList thus list1 and list2 
*   cannot be deleted after the join
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void JoinList(OFFSETLIST *list1,  /* the first list to be joined */
              OFFSETLIST *list2,  /* the second list to be joined */
              OFFSETLIST **newList      /* the result list of join     */
	      ) {
  OFFSETLIST *currNode1,
    *currNode2;
  OFFSETLIST *currNode3,
    *nextNode;

  if (*newList != NULL){
    printf("JoinList : newList is not empty \n");
    return;
  }

  if (list1 == NULL) {
    *newList = list2;
    return;
  } 
  else if (list2 == NULL) {
    *newList = list1;
    return;
  }
 
  currNode1 = list1;
  currNode2 = list2;
  while (currNode1 && currNode2){
    if (currNode1->Offset <= currNode2->Offset){
      nextNode = currNode1;
      currNode1 = currNode1->next;
    } else {
      nextNode = currNode2;
      currNode2 = currNode2->next;
    }
    if (*newList == NULL){
      *newList = nextNode;
      currNode3 = nextNode;
      currNode3->next = NULL;
    } else {
      currNode3->next = nextNode;
      currNode3 = currNode3->next;
      currNode3->next = NULL;
    }
  }

  if (currNode1){
    while (currNode1){
      currNode3->next = currNode1;
      currNode1 = currNode1->next;
      currNode3 = currNode3->next;
      currNode3->next = NULL;
    }
  } else if (currNode2) {
    while (currNode2){
      currNode3->next = currNode2;
      currNode2 = currNode2->next;
      currNode3 = currNode3->next;
      currNode3->next = NULL;
    }
  }
}

/***************************************************************************

* InsertList():
*   Routine to insert a node into the OFFSETLIST list preserving an increasing
*   order in the list
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   Memory space will be allocated for the new node
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void InsertList(OFFSETLIST **list, /* list where the new node to be inserted */
                const long offset  /* the offset that will form the new node */
		) {
  OFFSETLIST *currNode = NULL,
    *preNode = NULL,
    *newNode = NULL;

  newNode = (OFFSETLIST *) malloc(sizeof(OFFSETLIST));
  newNode->Offset = offset;
  newNode->next = NULL;

  if (*list == NULL){
    *list = newNode;
    newNode->next = NULL;
  }else{
    currNode = *list;
    preNode = NULL;
    while (currNode && (newNode->Offset > currNode->Offset)){
      preNode = currNode;
      currNode = currNode->next;
    }
    if (preNode){  
      preNode->next = newNode;
      newNode->next = currNode;
    }else{  /* insert at the beginning of the list */
      *list = newNode;
      newNode->next = currNode;
    }
  }
}

/***************************************************************************

* PrintList():
*   Routine to display the list for debugging
*
* IMPORTS:
*   None.
*
* RETURNS:
*   None.
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   None.
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void PrintList(const OFFSETLIST* list
	       ) {
  const OFFSETLIST *currNode;
  int nextLine = 0;
  int total = 0;
  
  if (list == NULL) {
    printf("PrintList : list is empty\n");
  } else {
    currNode = list;
    while(currNode){
      if (nextLine >= 8){
        printf("\n");
        nextLine = 0;
      }
      printf("%d->",currNode->Offset);
      currNode = currNode->next;
      nextLine++;
      total++;
    }
    printf("END\n");
    printf("Number of offsets is %d\n", total);
  }
}

void PrintList_v2(const OFFSETLIST* list
		  ) {
  const OFFSETLIST *currNode;
  int total = 0;
  
  if (list == NULL) {
    printf("0\n");
  } else {
    currNode = list;
    while(currNode){
      currNode = currNode->next;
      total++;
    }
    printf("%d\n", total);
  }
}

/***************************************************************************

* IntersectList():
*   Routine to take the intersection of the two lists
*
* IMPORTS:
*   None.
*
* RETURNS:
*   The intersection of the two lists, which is still in increasing order
*
* ERROR RETURNS:
*   None.
*
* SIDE EFFECTS:
*   Memory spaces are allocated to the new list
*
* WARNINGS:
*   None.
*
* BUGS:
*   Not discovered yet
***************************************************************************/
void IntersectList(const OFFSETLIST *list1,
                   const OFFSETLIST *list2,
                   OFFSETLIST **newList
		   ) {
  const OFFSETLIST *currNode1,
    *currNode2;
  OFFSETLIST *newNode, 
    *currNode3;

  currNode1 = list1;
  currNode2 = list2;
  while (currNode1 && currNode2){
    if (currNode1->Offset < currNode2->Offset){
      currNode1 = currNode1->next;
    } else {
      if (currNode1->Offset > currNode2->Offset){
        currNode2 = currNode2->next;
      } else {
        newNode = (OFFSETLIST *) malloc(sizeof(OFFSETLIST));
        newNode->Offset = currNode1->Offset;
        newNode->next = NULL;
        currNode1 = currNode1->next;
        currNode2 = currNode2->next;
        if (*newList == NULL){
          *newList = newNode;
          currNode3 = *newList;
        } else {
          currNode3->next = newNode;
          currNode3 = currNode3->next;
        }
      }
    }
  }
}
