/* Version 1.1 (Beta Test)
 * Copyright (c) Rui Zhang, 2005-2006
 *
 * Permission is hereby given for the use of the code subject to the
 *  following conditions:
 *  - The code will not be sold for profit without explicit written
 *    permission from Rui Zhang.
 *  - This copyright notice and author information will not be altered
 *  - All bug fixes will be returned to the rui@csse.unimelb.edu.au for
 *    inclusion in future releases
 */


/*************************************************************************
 *
 * File : operations.c
 *
 * Modules:
 * Operations of the GiMP
 *	BuildIndex()
 *	Key()
 *	GiMP_Insert()
 *	GiMP_Delete()
 *	RangeSearch()
 *	CompareInterval()
 *	CheckRange()
 *	kNNSearch()
 *	PrintAnswers()
 *	CompareAnswer()
 *
 * Version 1.1
 * Coded by: Rui Zhang
 *************************************************************************/

#include "btree.h"
#include "customize.h"
#include "operations.h"

void BuildIndex (BTREE *btree, char *DataFileName, FILE *fp_data, char *IndexFileName, FILE *fp_index, FILE *fp_btreeinfo)
{
  KEYTYPE key;
  RECORD_TYPE record;
  long rid;
  int i;

  if ( (fp_data = fopen (DataFileName, "r")) == NULL) //open dataFile
    {
      printf ("Can not open file %s\n", DataFileName);
      exit(0);
    }
  if ( (fp_index = fopen (IndexFileName, "w+")) == NULL) //open indexFile
    {
      printf ("Can not open file %s\n", IndexFileName);
      exit(0);
    }
  if ( (fp_btreeinfo = fopen ("btreeinfo", "w+")) == NULL) //open btreeInfo
    {
      printf ("Can not open file btreeinfo\n");
      exit(0);
    }

  //open a B+ tree, function defined in btree.c
  BT_Open (btree, fp_index, -1, -1);

  fseek (fp_data, 0L, SEEK_SET);

  rid=0;

  //insert record one by one
  while (fread (&record, sizeof (RECORD_TYPE), 1, fp_data))
    {
      GiMP_Insert (btree, rid, &record);
      rid++;
    }

  if (compactrate > 0)
    compact (btree);//compact nodes to improve utilization
  refine (btree);//adding Prev point into leaf nodes

  fprintf (fp_btreeinfo, "%ld\n", btree->Root);
  fprintf (fp_btreeinfo, "%ld\n", btree->Leaf_Head);

  for (i = 0; i < NumRef; i++)
    {
      fprintf (fp_btreeinfo, "%f ", min_r[i]);
      fprintf (fp_btreeinfo, "%f\n", max_r[i]);   
    }

  fclose (fp_btreeinfo);
  fclose (fp_data);
  BT_Close (btree);

  printf ("\nBuilding index is finished.\n");
}


KEYTYPE	Key (RECORD_TYPE* pRcd, int *refnum)
{
  return Base (pRcd) + Distance (pRcd, Reference (pRcd, refnum));
}

int GiMP_Insert (BTREE *btree, const long rid, RECORD_TYPE *pRcd)
{
  KEYTYPE key;
  int i;
  key = Key(pRcd, &i);

  //collect information on max/min keys when having multiple reference points
  if (key > max_r[i])
    max_r[i] = key; //max_r[i] stores the max key in partition i
  if (key < min_r[i])
    min_r[i] = key; //min_r[i] stores the min key in partition i

  //insert a new record to the B+ tree, function defined in btree.c
  return BT_Insert (btree, rid, key, pRcd);
}

int GiMP_Delete (BTREE *btree, RECORD_TYPE *pRcd)
{
  KEYTYPE key;
  int i;
  key = Key(pRcd, &i);

  //delete all entries which match a key and the record, function defined in btree.c
  return BT_DeleteRecord (btree, pRcd, key);
}

RecordListNode* RangeSearch (RANGE_TYPE *pRg, BTREE *btree)
{
  int i, ni;
  IntervalListNode *pILN, *CurILN, *oldILN;
  Interval *intvlarray;
  RecordListNode *cddhead; //the head does not contain a point

  //function defined in customize.c
  CurILN = pILN = MapRange(pRg);

  //get the number of intervals
  for (i = 0; (CurILN = CurILN->next) != NULL; i++);
  ni = i + 1;

  //put the intervals in an array and sort them according to lower bounds
  if ( (intvlarray = (Interval *) malloc (sizeof (Interval) * ni)) == NULL)
    {
      printf ("out of memory, use other sorting method!\n");
      exit(0);
    }

  //put the intervals in intvlarray
  for (i = 0, CurILN = pILN; CurILN != NULL; i++)
    {
      intvlarray[i].lower = CurILN->lower;
      intvlarray[i].higher = CurILN->higher;
      oldILN = CurILN;
      CurILN = CurILN->next;
      free (oldILN);
    }

  //sort the intervals
  qsort (intvlarray, ni, sizeof (Interval), CompareInterval);

  cddhead = (RecordListNode *) malloc (sizeof (RecordListNode));
  cddhead->next = NULL;

  //b+ tree range search
  //optimized for a set of intervals sorted by their lower bounds
  BT_RangeSearch_SortedIntervals (btree, cddhead, intvlarray, ni);

  CheckRange (cddhead, pRg);

  free (intvlarray);

  return cddhead;
}

int CompareInterval (const void *arg1, const void *arg2)
{
  if ( ((Interval *) arg1)->lower < ( (Interval *) arg2)->lower )
    return -1;
  else if ( ((Interval *) arg1)->lower > ((Interval *) arg2)->lower )
    return 1;
  else
    return 0;
}

//check each point in the list pRLN against the range pRg
//delete any that is outside the range
//refer to the paper for more details
void CheckRange (RecordListNode *pRLN,RANGE_TYPE *pRg)
{
  RecordListNode *pCurNode, *pTemp;
  int j;
  pCurNode = pRLN;
  while (pCurNode->next != NULL)
    {
      for (j = 0; j<D
	     && pRg->lower[j] <= pCurNode->next->rcd.fields[j]
	     && pCurNode->next->rcd.fields[j] <= pRg->higher[j]; j++);
      if (j == D)
	pCurNode = pCurNode->next;
      else
	{
	  pTemp = pCurNode->next;
	  pCurNode->next = pCurNode->next->next;
	  free(pTemp);
	}
    }
}

RecordListNode* kNNSearch (RECORD_TYPE *pRcd, BTREE *btree, KnnInfo *knninfo)
{
  IntervalListNode *pILN = NULL;
  DATA_TYPE rmin, rmax;
  RecordListNode *answerset, *head;
  int i;

  answerset = (RecordListNode *) malloc (sizeof (RecordListNode) * knninfo->k);
  rmax = r0;

  //increase rmin and rmax by dr every step
  do
    {
      knninfo->r = rmax;
      //function defined in customize.c
      pILN = MapAnnulus(pRcd, rmin, rmax, knninfo);
      rmin = rmax;
      rmax = rmin + dr;
      if (newqueryflag)
	newqueryflag = 0;
    }
  while (BT_RangeSearch_kNN(btree, pRcd, answerset, knninfo, pILN));

  for (i=0; i < knninfo->k - 1; i++)
    answerset[i].next = &(answerset[i+1]);
  answerset[knninfo->k - 1].next = NULL;

  head = (RecordListNode *) malloc (sizeof (RecordListNode));
  head->next = answerset;

  return head;
}

//extern KnnInfo knninfo;

void PrintAnswers(RecordListNode *pRcd, FILE *fp)
{
  RecordListNode *CurRcd, *answerarray;
  int i, numanswer;

#define SORT_BYRID
#ifdef SORT_BYRID

  //get the number of answers
  for (CurRcd = pRcd->next, numanswer = 0; CurRcd != NULL; numanswer++, CurRcd = CurRcd->next);

  //put records into an array
  answerarray = (RecordListNode *) malloc (sizeof (RecordListNode) * numanswer);

  for (CurRcd = pRcd->next, i = 0; CurRcd != NULL; i++, CurRcd = CurRcd->next)
    answerarray[i] = *CurRcd;

  //sort the answers in the array
  qsort (answerarray, numanswer, sizeof (RecordListNode), CompareAnswer);

  for (i = 0; i < numanswer; i++)
    {
      fprintf (fp, "%ld", answerarray[i].rid);
      //for(i=0;i<D;i++) fprintf(fp, "%f ", (float) answerarray[i].rcd.fields[i]);
      fprintf (fp, "\n");
    }

#else

  i=0;
  for (CurRcd = pRcd->next; CurRcd != NULL; CurRcd = CurRcd->next)
    {
      fprintf(fp, "%ld", CurRcd->rid);
      //for(i=0;i<D;i++) fprintf(fp, "%f ", (float) CurRcd->rcd.fields[i]);
      fprintf(fp, "\n");
    }

#endif
}

int CompareAnswer (const void *arg1, const void *arg2)
{
  if ( ( (RecordListNode *) arg1)->rid < ( (RecordListNode *) arg2)->rid)
    return -1;
  else if ( ( (RecordListNode *) arg1)->rid > ( (RecordListNode *) arg2)->rid)
    return 1;
  else return 0;
}
